#import <Foundation/Foundation.h>

@class NSBundle;
@class NSString;
@class NSArray;
@class NSDictionary;
@class LKAST;
@class LKModule;
@class LKMethod;
@protocol LKCodeGenerator;

typedef enum
{
	LKDebuggingDisabled = 0,
	LKDebuggingEnabled = 1
} LKDebuggingMode;

/**
 * Type of a block to be called when the runtime tries to resolve a class that
 * doesn't yet exist.
 */
typedef void(^LKClassLoader)(NSString*);

/**
 * All languages must use a parser conforming to this protocol.
 */
@protocol LKParser <NSObject>
/**
 * Returns a module AST constructed by parsing the specified source.
 * The result is nil if parsing failed.
 */
- (LKModule*) parseString:(NSString*)source;
/**
 * Returns a method AST constructed by parsing the specified source, or nil if
 * parsing failed or method parsing is not supported by this language.
 */
- (LKMethod*) parseMethod:(NSString*)source;
@end

@class LKCompiler;

@protocol LKCompilerDelegate<NSObject>
/**
 * Indicates that the specified compiler has encountered a recoverable problem.
 * The delegate should return YES if the compiler should attempt to continue.
 * Extra information about this warning may be found in the info dictionary.
 * All warnings must contain a human readable description identified by the key
 * kLKHumanReadableDesciption, specific warnings may contain other information.
 * See the LKCompilerErrors.h file for more information.
 */
- (BOOL)compiler: (LKCompiler*)aCompiler
generatedWarning: (NSString*)aWarning
         details: (NSDictionary*)info;
/**
 * Indicates that the specified compiler has encountered an unrecoverable
 * problem.  Extra information about this error  may be found in the info
 * dictionary.  All errors must contain a human readable description
 * identified by the key kLKHumanReadableDescription, specific errors may
 * contain other information.  See the LKCompilerErrors.h file for more
 * information.
 *
 * In normal circumstances, errors are unrecoverable.  A delegate may perform
 * some manipulation on the AST to eliminate the error, however.  In this case,
 * the delegate should return YES, causing the semantic analysis to be retried.
 */
- (BOOL)compiler: (LKCompiler*)aCompiler
  generatedError: (NSString*)anError
         details: (NSDictionary*)info;
@end

/**
 * Abstract class implementing a dynamic language compiler.  Concrete
 * subclasses provide implementations for specific languages.
 */
@interface LKCompiler : NSObject 
{
	id<LKCompilerDelegate> delegate;
}
@property(nonatomic, retain) NSMutableArray *transforms;
/**
 * Sets the default delegate that will be used by new compiler instances.  If
 * this is not called then a default delegate will be installed that just logs
 * errors and warnings to the console.
 */
+ (void)setDefaultDelegate: (id<LKCompilerDelegate>)aDelegate;
/**
 * Sets the delegate for this compiler.  The delegate is responsible for
 * handling errors and warnings that occur during the semantic analysis of the
 * program.
 */
- (void)setDelegate: (id<LKCompilerDelegate>)aDelegate;
/**
 * Adds a new class loader.  These will be called in the order that they
 * appear.
 */
+ (void)addClassLoader: (LKClassLoader)aBlock;
/**
 * Returns a new autoreleased compiler for this language.
 */
+ (LKCompiler*) compiler;
/**
 * Returns whether we are in developer mode.  In this mode, all unrecognised
 * symbols are assumed to be classes.
 */
+ (BOOL)inDevMode;
/**
 * Specifies whether the compiler should run in debug mode.  Enabling this will
 * spam stderr with a huge amount of debugging information.  Note that this is
 * a global setting and will apply to all compilers.
 */
+ (void) setDebugMode:(LKDebuggingMode)aFlag;
/**
 * Compiles and loads the specified source code.  Returns the AST on success,
 * nil on failure.
 */
- (LKAST*) compileString:(NSString*)source;
/**
 * Compiles the specified source code to LLVM bitcode.  This can then be
 * optimised with the LLVM opt utility and converted to object code with llc.
 * Returns the AST on success, nil on failure.
 */
//- (LKAST*) compileString:(NSString*)source outputDir:(NSString*)dir;
/**
 * Compiles the specified source code using the given code generator. Returns
 * the AST on success, nil on failure.
 */
- (LKAST*) compileString:(NSString*)source withGenerator:(id<LKCodeGenerator>)cg;
/**
 * Compiles and loads a method on the class with the specified name.
 */
- (BOOL) compileMethod:(NSString*)source onClassNamed:(NSString*)name;
/**
 * Compiles a method to LLVM bitcode for the class with the specified name.
 */
- (BOOL) compileMethod:(NSString*)source 
          onClassNamed:(NSString*)name 
                output:(NSString*)bitcode;
/**
 * Compiles a method on the class with the specified name using the given code
 * generator.
 */
- (BOOL) compileMethod:(NSString*)source
          onClassNamed:(NSString*)name
         withGenerator:(id<LKCodeGenerator>)cg;
/**
 * Load a framework with the specified name.
 */
+ (BOOL) loadFrameworkNamed:(NSString*)framework;
/**
 * Compiles and loads a file from the specified bundle.
 * The file name extension will be used to select a suitable compiler.
 */
+ (BOOL) loadScriptNamed: (NSString*)fileName fromBundle: (NSBundle*)bundle;
/**
 * Compiles and loads a file from the specified bundle.
 * Omit the extension in the name paramater.
 */
- (BOOL) loadScriptNamed: (NSString*)fileName fromBundle: (NSBundle*)bundle;
/**
 * Compiles and loads a file from the application bundle.
 * The file name extension will be used to select a suitable compiler.
 */
+ (BOOL) loadApplicationScriptNamed:(NSString*)fileName;
/**
 * Compiles and loads a file from the application bundle.
 * Omit the extension in the name paramater.
 */
- (BOOL) loadApplicationScriptNamed:(NSString*)name;
/**
 * Loads all scripts for all known languages from the specified bundle.
 */
+ (BOOL) loadScriptsFromBundle:(NSBundle*) aBundle;
/**
 * Loads all scripts written in this language from the specified bundle.
 */
- (BOOL) loadScriptsFromBundle:(NSBundle*) aBundle;
/**
 * Loads all scripts for all known languages from the application bundle.
 */
+ (BOOL) loadAllScriptsForApplication;
/**
 * Loads all scripts written in this language from the application bundle.
 */
- (BOOL) loadAllScriptsForApplication;

/**
 * Loads a bundle containing an LKInfo.plist file.  This should contain a
 * Source key giving an array of source files in the order in which they should
 * be compiled, a Frameworks key giving an array of frameworks and a
 * Classes key for declaring classes being compile in other to resolve symbols
 * correctly
 */
+ (BOOL) compileLanguageKitBundle: (NSBundle*)bundle output: (NSString*)bitcode;
/**
 * Loads a bundle containing an LKInfo.plist file.  This should contain a
 * Source key giving an array of source files in the order in which they should
 * be compiled, a Frameworks key giving an array of frameworks, a Classes key for
 * declaring classes being load in other to resolve symbols correctly and a
 * PrincipalClass key for the class that should be instantiated when it is
 * loaded.
 *
 * Returns (Class)-1 in case of error, or the principal class if loading
 * succeeds (Nil if no principal class is specified).
 */
+ (Class) loadLanguageKitBundle:(NSBundle*)bundle;
/**
 * Loads all of the LanguageKit plugins for the current application.
 */
+ (BOOL) loadAllPlugInsForApplication;
/**
 * Attempts to parse a C header.  Returns YES on success, NO on failure.
 * Failure can be caused by an inability to locate the header or by
 * SourceCodeKit not being available.
 */
+ (BOOL) loadHeader: (NSString*)aHeader;
/**
 * Attempts to load a library from an absolute path or one of the standard
 * library locations.
 */
+ (BOOL) loadLibrary: (NSString*)aLibrary;
/**
 * <override-subclass /> 
 * Returns the extension used for scripts in this language.
 *
 * Implemented by subclasses.
 */
+ (NSString*) fileExtension;
/**
 * <override-subclass /> 
 * Returns the name of the language supported by this compiler.
 *
 * Implemented by subclasses.
 */
+ (NSString*) languageName;
/**
 * Returns the languages for which compilers are currently supported.
 */
+ (NSArray*) supportedLanguageNames;
/**
 *	Returns the compiler for a named language. 
 */
+ (Class) compilerForLanguage:(NSString*) aLanguage;
/**
 * Returns the compiler for files with a given extension.
 */
+ (Class) compilerClassForFileExtension:(NSString*) anExtension;
/**
 * <override-subclass /> 
 * Returns the parser used by this language.
 */
+ (Class) parserClass;
/**
 * Reports the given warning and returns YES if recovery should proceed.
 */
+ (BOOL)reportWarning: (NSString*)aWarning
              details: (NSDictionary*)info;
/**
 * Reports the given error and returns YES if recovery should proceed.
 */
+ (BOOL)reportError: (NSString*)aWarning
            details: (NSDictionary*)info;

+ (LKAST*) parseScript: (NSString*)script forFileExtension: (NSString*)extension;
@end

@interface LKCompiler (JTL)
+ (void) justTooLateCompileBundle: (NSBundle*)aBundle;
/**
 * Link LKModule into a library. 
 */
+ (NSString*)linkBitcodeFiles: (NSMutableArray*) files outputDir: (NSString*)dir outputFile: (NSString*)fileName;
@end
